"""Base app controller module."""
from csv import DictReader
from datetime import datetime
from pathlib import Path
from shutil import copyfile
from sqlite3 import Cursor, Error, OperationalError
from string import Formatter
from typing import Dict, List, Optional
from xml.etree.ElementTree import Element, parse as xml_parse  # noqa: S405

from cement import Controller, ex
from cement.utils.version import get_version_banner

from jinja2 import Environment, FileSystemLoader

from tabulate import tabulate

from ..core.version import get_version

VERSION_BANNER = """
Yager %s
%s
""" % (
    get_version(),
    get_version_banner(),
)


class Base(Controller):
    """ Class implementing base app controller."""

    class Meta:
        """Controller meta-data."""

        label = "base"

        description = "Yet Another GEneric Reporter tool for parsing of XML data into an SQLite database and subsequent universal reporting based on SQL queries and Jinja2 templates."  # noqa: E501
        epilog = "Usage: yager {sub-command} {options}"

        arguments = [
            (["-v", "--version"], {"action": "version", "version": VERSION_BANNER}),
        ]

    def _query_db(self, sql_query: str) -> Optional[Cursor]:
        """Query database.

        Helper to execute `sql_query` against the app database.

        Parameters
        ----------
        sql_query
            SQLite3 query to be executed.

        Returns
        -------
        :obj:`~typing.Optional` [:obj:`~sqlite3.Cursor`]
            If successful, returns :obj:`~sqlite3.Cursor` with results.
            Otherwise returns ``None``.
        """
        # Stop processing, if DB is not available
        if self.app.db_cursor is None:
            self.app.log.error("No database connection")
            return

        self.app.log.debug("Executing query '{}'".format(sql_query))

        response: Optional[Cursor] = None

        try:
            response = self.app.db_cursor.execute(sql_query)

        except OperationalError as e:
            self.app.log.error("sqlite3.OperationalError: {}".format(str(e)))

        except Error as e:
            self.app.log.error("Unexpected Error: {}".format(str(e)))

            if self.app.debug:
                import traceback

                traceback.print_exc()

        else:
            return response

    @ex(
        help="execute a query against database",
        arguments=[
            (
                ["--output", "-o"],
                {
                    "help": "defines how to format the output (choose from 'csv' or 'table'; default: 'table')",  # noqa: E501
                    "action": "store",
                    "metavar": "FORMAT",
                    "required": False,
                    "dest": "output_format",
                    "choices": ["csv", "table"],
                    "default": "table",
                },
            ),
            (["sql_query"], {"help": "query to be executed", "metavar": "QUERY"},),
        ],
    )
    def query(self) -> None:
        """Execute a query against the app database."""
        # Stop processing, if DB is not available
        if self.app.db_cursor is None:
            self.app.log.error("No database connection")
            return

        # get required params from CLI
        out_format: str = self.app.pargs.output_format
        sql_query: str = self.app.pargs.sql_query

        # init local vars
        sql_header: List[str] = []
        sql_response: Optional[Cursor] = None
        sql_data: List[str] = []

        sql_response = self._query_db(sql_query)
        if sql_response:
            sql_header = list(map(lambda x: x[0], self.app.db_cursor.description))
            sql_data = sql_response.fetchall()

            if out_format == "table":
                print(tabulate(sql_data, headers=sql_header))

            elif out_format == "csv":
                self.app.render(
                    {"records": sql_data, "header": sql_header}, "output_csv.j2",
                )

            else:
                self.app.log.error("Unknown output format {}".format(out_format))

    @ex(
        help="execute a pre-configured report",
        arguments=[
            (
                ["--param", "-p"],
                {
                    "help": "NAME=VALUE pair defining a global query parameter \
                        to be used for each query in the template \
                        (could be repetated)",
                    "action": "append",
                    "metavar": "PARAM",
                    "dest": "param_list",
                },
            ),
            (["report_name"], {"help": "report to be executed", "metavar": "NAME"},),
        ],
    )
    def report(self) -> None:
        """Execute a pre-configured report."""
        # Stop processing, if DB is not available
        if self.app.db_cursor is None:
            self.app.log.error("No database connection")
            return

        # get required params from app config
        all_report_configs: List[Dict] = self.app.config.get("yager", "reports")
        all_report_names: List[str] = list(map(lambda x: x["name"], all_report_configs))

        # get required params from CLI
        query_params_list: List = self.app.pargs.param_list

        report_name: str = self.app.pargs.report_name
        if report_name not in all_report_names:
            self.app.log.error("Unknown report name '{}'".format(report_name))
            self.app.log.debug(
                "Available reports are: '{}'".format("','".join(all_report_names))
            )

            return

        # init local vars
        report_config: Dict = all_report_configs[all_report_names.index(report_name)]
        template_params: List = report_config["template_params"]
        template_input_data: Dict = {}

        self.app.log.info(
            "Executing report template '{}'".format(report_config["name"])
        )

        for param in template_params:
            # expand possible query parameters
            sql_query_parametrized: str = param["query"]
            self.app.log.debug("Preparing query '{}'".format(sql_query_parametrized))

            query_keys = [
                t[1]
                for t in Formatter().parse(sql_query_parametrized)
                if t[1] is not None
            ]
            self.app.log.debug("Tokens used by query: '{}'".format(query_keys))

            # convert CLI list of query params to a dict
            quey_params_dict: Dict = {
                query_param.split("=")[0]: query_param.split("=")[1]
                for query_param in query_params_list
            }
            self.app.log.debug("Token values from CLI: '{}'".format(quey_params_dict))

            sql_query: str = sql_query_parametrized.format(**quey_params_dict)

            sql_result: Optional[Cursor] = None
            var_mapping: Dict = param["var_mapping"]

            sql_result = self._query_db(sql_query)
            if sql_result:
                data_tuples: List = sql_result.fetchall()
                data_header: List = [h[0] for h in sql_result.description]
                data_dicts: List[Dict] = [
                    {key: value for key, value in zip(data_header, row)}
                    for row in data_tuples
                ]

                self.app.log.debug("Resulting data '{}'".format(data_dicts))

                for var_key, var_value in var_mapping.items():
                    if var_value == "*":
                        template_input_data.update({var_key: data_dicts})
                    else:
                        template_input_data.update(
                            {var_key: data_dicts[0].get(var_value, "None")}
                        )

        jinja_env: Environment = Environment(  # noqa: S701
            loader=FileSystemLoader(
                self.app.config.get("yager", "data")["template_dir"]
            ),
            autoescape=False,
        )
        template = jinja_env.get_template(report_config["template_file"])
        output_from_parsed_template = template.render(template_input_data)
        print(output_from_parsed_template)

        self.app.log.info("Finished report template '{}'".format(report_config["name"]))

    @ex(
        help="generate database from configured data sources",
        arguments=[
            (
                ["--file", "-f"],
                {
                    "help": "path to XML file with data \
                        (could be repetated)",
                    "action": "append",
                    "metavar": "PATH",
                    "dest": "xml_list",
                },
            ),
        ],
    )
    def refresh_db(self) -> None:
        """Generate database from configured layout and data sources."""
        # Stop processing, if DB is not available
        if self.app.db_cursor is None:
            self.app.log.error("No database connection")
            return

        # get required params from app config
        all_data_configs: List[Dict] = self.app.config.get("yager", "data")

        # get required params from CLI
        xml_files: str = self.app.pargs.xml_list

        # init local vars
        database_query: Optional[Cursor] = None

        # make backup of main DB
        database_query = self._query_db("PRAGMA database_list;")
        if database_query:
            db_list: List = database_query.fetchall()
            db_name_dict: Dict = dict(map(lambda x: (x[1], x[2]), db_list))
            main_db_path: str = Path(db_name_dict["main"])
            db_backup_name: str = "{}.{}.bak".format(
                main_db_path.name, int(datetime.utcnow().timestamp())
            )

            self.app.log.info(
                "Backup DB '{}' as '{}'".format(main_db_path, db_backup_name)
            )

            copyfile(
                main_db_path, main_db_path.with_name(db_backup_name),
            )

        # delete all tables  from existing DB but leave excluded ones
        database_query = self._query_db(
            "SELECT name FROM sqlite_master WHERE type == 'table';"
        )
        if database_query:
            table_list: List[str] = database_query.fetchall()

            for record_tuple in table_list:
                table = record_tuple[0]
                if table not in all_data_configs["exclude_from_refresh"]:
                    self.app.log.info("Deleting table '{}'".format(table))
                    self._query_db("DROP TABLE {};".format(table))
                else:
                    self.app.log.info("Leaving table '{}' untouched".format(table))

        # create table layout
        for table in all_data_configs["layout"]:
            table_name: str = table["name"]
            table_columns: str = table["columns"].replace("\n", "")
            data_source_type: str = table["data_source"].split(":")[0]
            data_source: str = table["data_source"].split(":")[1]

            self.app.log.info("Creating table '{}'".format(table_name))
            database_query = self._query_db(
                "CREATE TABLE IF NOT EXISTS {} ({})".format(table_name, table_columns)
            )

            self.app.log.debug(
                "Using data source '{}:{}'".format(data_source_type, data_source)
            )
            if data_source_type == "csv":
                data_source_path: Path = Path(data_source)
                row_count: int = 0

                self.app.log.info(
                    "Inserting data from CSV file '{}'".format(data_source_path)
                )

                # read data from CSV
                with open(data_source_path, "r") as csv_file:
                    csv_data = DictReader(csv_file)

                    self._query_db("BEGIN TRANSACTION")

                    for row in csv_data:
                        self._query_db(
                            "INSERT OR IGNORE INTO {} ({}) VALUES ('{}')".format(
                                table_name,
                                ",".join(row.keys()),
                                "','".join(row.values()),
                            )
                        )

                        row_count += 1

                        # Commit to DB every 100 records
                        if row_count % 100 == 0:
                            self._query_db("COMMIT")

                            self.app.log.info(
                                "Inserted {} records...".format(row_count)
                            )

                            self._query_db("BEGIN TRANSACTION")

                    self._query_db("COMMIT")

                self.app.log.info("Total records inserted: {}".format(row_count))

            elif data_source_type == "xml":
                if xml_files:
                    for file_path in xml_files:
                        element_count: int = 0

                        self.app.log.info(
                            "Inserting data from XML file '{}'".format(file_path)
                        )

                        xml_root: Element = xml_parse(file_path).getroot()  # noqa: S314

                        self.app.log.info(
                            "Finding all elements matching XPath '{}'".format(
                                data_source
                            )
                        )

                        xml_data: List[Element] = xml_root.findall(data_source)
                        self.app.log.info("Found {} elements".format(len(xml_data)))

                        self._query_db("BEGIN TRANSACTION")

                        for element in xml_data:
                            input_map: Dict[str] = table["input_map"]
                            input_map_parametrized: Dict[str] = table.get(
                                "input_map_parametrized", {}
                            )
                            row: Dict[str] = {}

                            # Expand straightforward input paramenters into values
                            for input_key, input_value_map in input_map.items():
                                element_mapped: Optional[Element] = element.find(
                                    input_value_map
                                )

                                if element_mapped is not None:
                                    input_value: str = element_mapped.text
                                else:
                                    input_value = "undefined"

                                row.update({input_key: input_value})

                            # Expand parametrized input paramenters into values
                            for (
                                input_key,
                                input_value_map,
                            ) in input_map_parametrized.items():
                                input_value_map_expanded = input_value_map.format(**row)

                                element_mapped: Optional[Element] = xml_root.find(
                                    input_value_map_expanded
                                )

                                if element_mapped is not None:
                                    input_value: str = element_mapped.text
                                else:
                                    input_value = "undefined"

                                row.update({input_key: input_value})

                            self._query_db(
                                "INSERT OR IGNORE INTO {} ({}) VALUES ('{}')".format(
                                    table_name,
                                    ",".join(row.keys()),
                                    "','".join(row.values()),
                                )
                            )

                            element_count += 1

                            # Commit to DB every 100 records
                            if element_count % 100 == 0:
                                self._query_db("COMMIT")

                                self.app.log.info(
                                    "Inserted {} records...".format(element_count)
                                )

                                self._query_db("BEGIN TRANSACTION")

                        self._query_db("COMMIT")

                        self.app.log.info(
                            "Total records inserted: {}".format(element_count)
                        )

                else:
                    self.app.log.error("No XML files specified for data input")

            else:
                self.app.log.error(
                    "Unknown data source type '{}'".format(data_source_type)
                )



# -*- coding: utf-8 -*-
"""Framework hooks module."""
from sqlite3 import Connection, Cursor, Error, OperationalError, connect
from typing import Dict

from cement import App

from .version import get_version


def load_db(app: App) -> None:
    """Extend app with SQLite3 database.

    Reads data from configured SQLite3 URI.

    Parameters
    ----------
    app
        Cement Framework application object.
    """
    app.log.debug("Extending app object with database connection")

    data_config: Dict[str, str] = app.config.get("yager", "data")
    if data_config:
        db_uri: str = data_config.get("db_uri", "file:/home/user/.yager/data/sqlite.db")

    app.log.debug("Using database at '{}'".format(db_uri))  # noqa: G001

    try:
        db_connection: Connection = connect(db_uri)
    except OperationalError as e:
        app.extend("db_cursor", None)

        app.log.error("OperationalError: {}".format(str(e)))  # noqa: G001

        if app.debug:
            import traceback

            traceback.print_exc()
    else:
        try:
            db_cursor: Cursor = db_connection.cursor()
        except Error as e:
            app.log.error("Error: {}".format(str(e)))  # noqa: G001

            if app.debug:
                import traceback

                traceback.print_exc()
        else:
            app.extend("db_cursor", db_cursor)


def log_app_version(app: App) -> None:
    """Log the version of the app.

    Parameters
    ----------
    app
        Cement Framework application object.
    """
    app.log.info("Yager version {}".format(get_version()))  # noqa: G001



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yager', 'yager.controllers', 'yager.core', 'yager.ext', 'yager.templates']

package_data = \
{'': ['*']}

install_requires = \
['cement>=3.0.4,<4.0.0',
 'colorlog>=4.2.1,<5.0.0',
 'jinja2>=2.11.2,<3.0.0',
 'pyyaml>=5.3.1,<6.0.0',
 'tabulate>=0.8.7,<0.9.0',
 'typing>=3.7.4,<4.0.0']

extras_require = \
{'code-format': ['black>=19.10b0,<20.0', 'blacken-docs>=1.7,<2.0'],
 'code-lint': ['flake8>=3.8,<4.0',
               'flake8-import-order>=0.18,<0.19',
               'flake8-bandit>=2.1,<3.0',
               'flake8-blind-except>=0.1,<0.2',
               'flake8-bugbear>=20.1,<21.0',
               'flake8-builtins>=1.5,<2.0',
               'flake8-docstrings>=1.5,<2.0',
               'flake8-logging-format>=0.6,<0.7',
               'flake8-mypy>=17.8,<18.0',
               'pep8-naming>=0.8,<0.9',
               'pygments>=2.6,<3.0'],
 'docs': ['recommonmark>=0.6.0,<0.7.0',
          'sphinx>=3.1,<4.0',
          'sphinx-rtd-theme>=0.5,<0.6',
          'sphinx-autodoc-typehints>=1.11,<2.0'],
 'test': ['pytest>=6.0,<7.0',
          'pytest-benchmark[aspect]>=3.2,<4.0',
          'pytest-cov>=2.10,<3.0',
          'pytest-instafail>=0.4,<0.5',
          'pytest-lazy-fixture>=0.6,<0.7',
          'pytest-random-order>=1.0,<2.0']}

entry_points = \
{'console_scripts': ['yager = yager.main:main']}

setup_kwargs = {
    'name': 'yager',
    'version': '0.1.0',
    'description': 'Yet Another GEneric Reporter tool for parsing of XML data into an SQLite database and subsequent universal reporting based on SQL queries and Jinja2 templates.',
    'long_description': '# Yager = Yet Another GEneric Reporter\n\n[![Python 3.7+](https://img.shields.io/badge/Python-3.7+-blue.svg)][PythonRef] [![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)][BlackRef] [![License: MIT](https://img.shields.io/badge/License-MIT-blue.svg)][MITRef]\n\n[PythonRef]: https://docs.python.org/3.7/\n[BlackRef]: https://github.com/ambv/black\n[MITRef]: https://opensource.org/licenses/MIT\n\n`yager` is Yet Another GEneric Reporter tool for parsing of XML data into an SQLite database and subsequent universal reporting based on SQL queries and Jinja2 templates.\n\n## Getting Started\n\n### Installing\n\n`yager` is distributed through the [Python Package Index][PyPIRef] as [yager][PyPIProjRef]. Run the following command to:\n\n[PyPIRef]: https://pypi.org\n[PyPIProjRef]:https://pypi.org/project/yager/\n\n* install a specific version\n\n    ```sh\n    pip install "yager==0.1"\n    ```\n\n* install the latest version\n\n    ```sh\n    pip install "yager"\n    ```\n\n* upgrade to the latest version\n\n    ```sh\n    pip install --upgrade "yager"\n    ```\n\n* install optional DEV dependencies like `pytest` framework and plugins necessary for performance and functional testing\n\n    ```sh\n    pip install "yager[test]"\n    ```\n\n### Configuring\n\n`yager` looks for a `YAML` configuration file in the following locations:\n\n* `/etc/yager/yager.yaml`\n* `~/.config/yager/yager.yaml`\n* `~/.yager/config/yager.yaml`\n* `~/.yager.yaml`\n\nBelow is the [example configuration file][yagerConfigRef] that parses XML data from [Qualys Cloud Agent API][QualysCloudAgentAPIRef] about Azure VMs and generates reports for each Azure Subscription.\n\n[yagerConfigRef]: config/etc/yager_example.yaml\n[QualysCloudAgentAPIRef]: https://www.qualys.com/docs/qualys-ca-api-user-guide.pdf\n\n```yaml\n### Yager Configuration Settings\n---\nyager:\n  # Toggle application level debug (does not toggle Cement framework debugging)\n  # debug: true\n\n  # Data location and layout\n  data:\n    # URI to the SQLite3 database with Qualys data\n    db_uri: "file:./instance/data/results.db"\n\n    # Tables exclude from deletion during `refresh-db`\n    exclude_from_refresh: [\n      "QualysKB",\n    ]\n\n    # List of table definitions\n    layout:\n        # Table name\n      - name: AzureSubscriptions\n        # Table columns described as an SQL statement;\n        # Used in `CREATE TABLE IF NOT EXISTS {name} ({columns})`\n        columns: |\n          id TEXT PRIMARY KEY,\n          name TEXT,\n          contactName TEXT,\n          contactEmail TEXT\n        # Source of data for the table in a form of `{type}:{locator}`\n        #   {type} could be `csv` or `xml`\n        #   {locator} is a file path for `csv` type or XPath for `xml`\n        #   XML files are provided with `--file path/to/file.xml` option\n        data_source: csv:./instance/subscriptions.csv\n\n      - name: HostAssets\n        columns: |\n          id INTEGER PRIMARY KEY,\n          name TEXT,\n          fqdn TEXT,\n          os TEXT,\n          firstSeen DATETIME,\n          lastUpdated DATETIME,\n          lastVulnScan DATETIME,\n          azureVmId TEXT,\n          azureRgName TEXT,\n          azurePublicIp TEXT,\n          azurePrivateIp TEXT,\n          azureSubscriptionId TEXT,\n          FOREIGN KEY (azureSubscriptionId)\n              REFERENCES AzureSubscriptions (id)\n        data_source: xml:.//HostAsset\n        # Mapping between table columns and XPath to text values\n        # of sub-elements for each element returned by `data_source` XPath\n        input_map:\n          id: id\n          name: name\n          fqdn: fqdn\n          os: os\n          firstSeen: created\n          lastUpdated: modified\n          lastVulnScan: lastVulnScan\n          azureVmId: ./sourceInfo/list/AzureAssetSourceSimple/vmId\n          azureRgName: ./sourceInfo/list/AzureAssetSourceSimple/resourceGroupName\n          azurePublicIp: ./sourceInfo/list/AzureAssetSourceSimple/publicIpAddress\n          azurePrivateIp: ./sourceInfo/list/AzureAssetSourceSimple/privateIpAddress\n          azureSubscriptionId: ./sourceInfo/list/AzureAssetSourceSimple/subscriptionId\n\n      - name: Vulns\n        columns: |\n          id INTEGER PRIMARY KEY,\n          qid INTEGER,\n          firstFound DATETIME,\n          lastFound DATETIME,\n          hostAssetsId INTEGER,\n          FOREIGN KEY (hostAssetsId)\n              REFERENCES HostAssets (id)\n        data_source: xml:.//HostAssetVuln\n        input_map:\n          id: hostInstanceVulnId\n          qid: qid\n          firstFound: firstFound\n          lastFound: lastFound\n        # Similar to `input_map` but mapping is parametrized with any {table_column}\n        # already defined in the `input_map` section. Prior to running XPath match,\n        # each {table_column} is expanded with a value already acquired for `input_map`.\n        input_map_parametrized:\n          hostAssetsId: ".//HostAssetVuln/[hostInstanceVulnId=\'{id}\']/.../.../.../id"\n\n    # Path to directory with Jinja2 templates for reports\n    template_dir: "./config/templates/"\n\n  # Templated reports\n  reports:\n    - name: "single_sub_md"\n      description: |\n        Report provides detailed Qualys vulnerability data for an individual Azure subscription.\n        It expects the following parameters to be provided in CLI:\n          * `sub_id` - Subscription ID from `AzureSubscriptions` table\n      template_file: "report_single_sub_md.j2"\n      template_params:\n        - query: |\n            -- Subscription details\n            SELECT\n              id,\n              name,\n              contactName,\n              contactEmail\n            FROM\n              AzureSubscriptions\n            WHERE\n              id == \'{sub_id}\'\n          var_mapping:\n            subscription_id: id\n            subscription_name: name\n            subscription_poc_name: contactName\n            subscription_poc_email: contactEmail\n\n        - query: |\n            -- Vuln details for all hosts in subscription\n            SELECT\n                HostAssets.name as host_name,\n                HostAssets.azureVmId as host_vm_id,\n                Vulns.firstFound as vuln_seen_first,\n                Vulns.lastFound as vuln_seen_last,\n                Vulns.qid as vuln_qid,\n                QualysKB.severity_level as vuln_level\n            FROM\n                HostAssets\n                INNER JOIN Vulns ON Vulns.hostAssetsId = HostAssets.id\n                INNER JOIN QualysKB ON QualysKB.qid = Vulns.qid\n            WHERE\n                HostAssets.azureSubscriptionId == \'{sub_id}\'\n                AND QualysKB.vuln_type != "Information Gathered";\n          var_mapping:\n            all_vuln_details: "*"\n\n        - query: |\n            -- Additional host details\n            SELECT\n                HostAssets.name as host_name,\n                HostAssets.fqdn as host_fqdn,\n                HostAssets.os as host_os,\n                HostAssets.lastVulnScan as host_last_scan,\n                HostAssets.azureRgName as host_rg,\n                HostAssets.azureVmId as host_vm_id,\n                HostAssets.azurePrivateIp as host_priv_ip,\n                HostAssets.azurePublicIp as host_pub_ip\n            FROM\n                HostAssets\n            WHERE\n                HostAssets.azureSubscriptionId == \'{sub_id}\'\n          var_mapping:\n            all_host_details: "*"\n\n        - query: |\n            -- Reference information from Qualys Knowledge Base\n            SELECT\n                QualysKB.qid as vuln_qid,\n                QualysKB.severity_level as vuln_level,\n                QualysKB.title as vuln_title,\n                QualysKB.diagnosis as vuln_details,\n                QualysKB.consequence as vuln_risk,\n                QualysKB.solution as vuln_solution\n            FROM\n                QualysKB\n          var_mapping:\n            kb_details: "*"\n\n# Logging configuration\nlog.colorlog:\n  # Whether or not to colorize the log file.\n  # colorize_file_log: false\n\n  # Whether or not to colorize the console log.\n  # colorize_console_log: true\n\n  # Where the log file lives (no log file by default)\n  # file: null\n\n  # The level for which to log.  One of: info, warning, error, fatal, debug\n  # level: INFO\n\n  # Whether or not to log to console\n  # to_console: true\n\n  # Whether or not to rotate the log file when it reaches `max_bytes`\n  # rotate: false\n\n  # Max size in bytes that a log file can grow until it is rotated.\n  # max_bytes: 512000\n\n  # The maximum number of log files to maintain when rotating\n  # max_files: 4\n\n```\n\n## Usage\n\n### Generic\n\n```term\nusage: yager [-h] [-d] [-q] [-v] {query,refresh-db,report} ...\n\nYet Another GEneric Reporter tool for parsing of XML data into an SQLite database and subsequent universal reporting based on SQL queries and Jinja2 templates.\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -d, --debug           full application debug mode\n  -q, --quiet           suppress all console output\n  -v, --version         show program\'s version number and exit\n\nsub-commands:\n  {query,refresh-db,report}\n    query               execute a query against database\n    refresh-db          generate database from configured data sources\n    report              execute a pre-configured report\n\nUsage: yager {sub-command} {options}\n```\n\n### Query\n\n```term\nusage: yager query [-h] [--output FORMAT] QUERY\n\npositional arguments:\n  QUERY                 query to be executed\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --output FORMAT, -o FORMAT\n                        defines how to format the output (choose from \'csv\' or\n                        \'table\'; default: \'table\')\n```\n\n### Refresh-db\n\n```term\nusage: yager refresh-db [-h] [--file PATH]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --file PATH, -f PATH  path to XML file with data (could be repetated)\n```\n\n### Report\n\n```term\nusage: yager report [-h] [--param PARAM] NAME\n\npositional arguments:\n  NAME                  report to be executed\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --param PARAM, -p PARAM\n                        NAME=VALUE pair defining a global query parameter to\n                        be used for each query in the template (could be\n                        repetated)\n```\n\n### Examples\n\nRefresh the database using the layout and data sources described in the YAML config.\n\n```sh\n$ yager refresh-db -f data.xml\n```\n\nGenerate report based on SQL queries and Jinja2 templates defined in the YAML config.\n\n```sh\n$ yager report TEMPLATE --param KEY1=VALUE1 --param KEY2=VALUE2\n```\n\n## Requirements\n\n* Python >= 3.7\n\n## Built using\n\n* [Cement Framework][CementRef] - CLI application framework\n\n[CementRef]: https://builtoncement.com/\n\n## Versioning\n\nWe use [Semantic Versioning Specification][SemVer] as a version numbering convention.\n\n[SemVer]: http://semver.org/\n\n## Release History\n\nFor the available versions, see the [tags on this repository][RepoTags]. Specific changes for each version are documented in [CHANGELOG.md][ChangelogRef].\n\nAlso, conventions for `git commit` messages are documented in [CONTRIBUTING.md][ContribRef].\n\n[RepoTags]: https://github.com/undp/yager/tags\n[ChangelogRef]: CHANGELOG.md\n[ContribRef]: CONTRIBUTING.md\n\n## Authors\n\n* **Oleksiy Kuzmenko** - [OK-UNDP@GitHub][OK-UNDP@GitHub] - *Initial design and implementation*\n\n[OK-UNDP@GitHub]: https://github.com/OK-UNDP\n\n## Acknowledgments\n\n* Hat tip to all individuals shaping design of this project by sharing their knowledge in articles, blogs and forums.\n\n## License\n\nUnless otherwise stated, all authors (see commit logs) release their work under the [MIT License][MITRef]. See [LICENSE.md][LicenseRef] for details.\n\n[LicenseRef]: LICENSE.md\n\n## Contributing\n\nThere are plenty of ways you could contribute to this project. Feel free to:\n\n* submit bug reports and feature requests\n* outline, fix and expand documentation\n* peer-review bug reports and pull requests\n* implement new features or fix bugs\n\nSee [CONTRIBUTING.md][ContribRef] for details on code formatting, linting and testing frameworks used by this project.\n',
    'author': 'Oleksiy Kuzmenko',
    'author_email': 'oleksiy.kuzmenko@undp.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/undp/yager',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



"""Main app module."""
from cement import App, TestApp, init_defaults
from cement.core.exc import CaughtSignal

from .controllers.base import Base
from .core.exc import YagerError
from .core.hooks import load_db, log_app_version
from .core.log import YagerLogHandler

# configuration defaults
CONFIG = init_defaults("yager", "yager.data", "yager.reports")


class Yager(App):
    """Yager primary application."""

    class Meta:
        """Application meta-data."""

        label = "yager"

        # configuration defaults
        config_defaults = CONFIG

        # call sys.exit() on close
        exit_on_close = True

        # register functions to hooks
        hooks = [
            ("post_setup", log_app_version),
            ("post_setup", load_db),
        ]

        # load additional framework extensions
        extensions = [
            "colorlog",
            "jinja2",
            "yaml",
        ]

        # configuration handler
        config_handler = "yaml"

        # configuration file suffix
        config_file_suffix = ".yaml"

        # set log handler
        log_handler = "colorlog_custom_format"

        # set the output handler
        output_handler = "jinja2"

        # register handlers
        handlers = [Base, YagerLogHandler]


class YagerTest(TestApp, Yager):
    """A sub-class of Yager that is better suited for testing."""

    class Meta:
        """Test application meta-data."""

        label = "yager"


def main():
    """App entry point."""
    with Yager() as app:
        try:
            app.run()

        except AssertionError as e:
            print("AssertionError > %s" % e.args[0])
            app.exit_code = 1

            if app.debug:
                import traceback

                traceback.print_exc()

        except YagerError as e:
            print("YagerError > %s" % e.args[0])
            app.exit_code = 1

            if app.debug:
                import traceback

                traceback.print_exc()

        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print("\n%s" % e)
            app.exit_code = 0


if __name__ == "__main__":
    main()



"""Module defines app test cases."""
from yager.main import YagerTest


def test_yager():
    """Test without any subcommands or arguments."""
    with YagerTest() as app:
        app.run()
        assert app.exit_code == 0  # noqa: S101


def test_yager_debug():
    """Test that debug mode is functional."""
    argv = ["--debug"]
    with YagerTest(argv=argv) as app:
        app.run()
        assert app.debug is True  # noqa: S101
